// ddstuff.cpp - a direct draw routines library
// from LaMothe's game engine library
 
// INCLUDES ///////////////////////////////////////////////////

#define WIN32_LEAN_AND_MEAN  // make sure all macros are included

#include <windows.h>         // include important windows stuff
#include <windowsx.h> 
#include <mmsystem.h>

#include <iostream.h>        // include important C/C++ stuff
#include <conio.h>
#include <stdlib.h>
#include <malloc.h>
#include <memory.h>
#include <string.h>
#include <stdarg.h>
#include <stdio.h>
#include <math.h>
#include <io.h>
#include <fcntl.h>
#include <fstream.h>

#include <ddraw.h>           // directX includes
#include "ddstuff.h"        // game library includes
#include "global.h"

// EXTERNALS //////////////////////////////////////////////////

extern HWND main_window_handle; // save the window handle
extern HINSTANCE main_instance; // save the instance

extern ofstream outFile; // debug info output file

// GLOBALS ////////////////////////////////////////////////////

LPDIRECTDRAW7         lpdd         = NULL;   // dd object
LPDIRECTDRAWSURFACE7  lpddsprimary = NULL;   // dd primary surface
LPDIRECTDRAWSURFACE7  lpddsback    = NULL;   // dd back surface
LPDIRECTDRAWPALETTE   lpddpal      = NULL;   // a pointer to the created dd palette
LPDIRECTDRAWCLIPPER   lpddclipper  = NULL;   // dd clipper
PALETTEENTRY          palette[256];          // color palette
PALETTEENTRY          save_palette[256];     // used to save palettes
DDSURFACEDESC2        ddsd;                  // a direct draw surface description struct
DDBLTFX               ddbltfx;               // used to fill
DDSCAPS2              ddscaps;               // a direct draw surface capabilities struct
HRESULT               ddrval;                // result back from dd calls
DWORD                 start_clock_count = 0; // used for timing

// these defined the general clipping rectangle
int min_clip_x = 0,                          // clipping rectangle 
    max_clip_x = SCREEN_WIDTH-1,
    min_clip_y = 0,
    max_clip_y = SCREEN_HEIGHT-1;

// these are overwritten globally by DD_Init()
int screen_width  = SCREEN_WIDTH,            // width of screen
    screen_height = SCREEN_HEIGHT,           // height of screen
    screen_bpp    = SCREEN_BPP;              // bits per pixel


// FUNCTIONS //////////////////////////////////////////////////

int DD_Init(int width, int height, int bpp)		// this function initializes directdraw
{
	int index; // looping variable
	int i;

	// create object and test for error
	if (DirectDrawCreateEx(NULL, (void **)&lpdd, IID_IDirectDraw7, NULL)!=DD_OK)
	   return(0);

	// set cooperation level to windowed mode normal
	if (lpdd->SetCooperativeLevel(main_window_handle,
			   DDSCL_ALLOWMODEX | DDSCL_FULLSCREEN | 
			   DDSCL_EXCLUSIVE | DDSCL_ALLOWREBOOT)!=DD_OK)
		return(0);

	if (lpdd->SetDisplayMode(width,height,bpp,0,0)!=DD_OK)	// set the display mode
		return(0);

	// set globals
	screen_height = height;
	screen_width  = width;
	screen_bpp    = bpp;

	// Create the primary surface
	memset(&ddsd,0,sizeof(ddsd));
	ddsd.dwSize = sizeof(ddsd);
	ddsd.dwFlags = DDSD_CAPS | DDSD_BACKBUFFERCOUNT;

	// we need to let dd know that we want a complex 
	// flippable surface structure, set flags for that
	ddsd.ddsCaps.dwCaps = DDSCAPS_PRIMARYSURFACE | DDSCAPS_FLIP | DDSCAPS_COMPLEX;
	ddsd.dwBackBufferCount = 1;						// set the backbuffer count to 1
	lpdd->CreateSurface(&ddsd,&lpddsprimary,NULL);	// create the primary surface

	// query for the backbuffer i.e the secondary surface
	ddscaps.dwCaps = DDSCAPS_BACKBUFFER;
	lpddsprimary->GetAttachedSurface(&ddscaps,&lpddsback);

	// create and attach palette
	// create palette data
	memset(palette, 0, 256 * sizeof(PALETTEENTRY));		// clear all entries defensive programming
	for (index = 0; index < 256; index++)				// create a R,G,B,GR gradient palette
	{
		// set each entry
		if (index < 64) 
			palette[index].peRed = index * 4; 
		else           // shades of green
		if (index >= 64 && index < 128) 
			palette[index].peGreen = (index - 64) * 4;
		else           // shades of blue
		if (index >= 128 && index < 192) 
		   palette[index].peBlue = (index - 128) * 4;
		else           // shades of grey
		if (index >= 192 && index < 256) 
			palette[index].peRed = palette[index].peGreen =
							palette[index].peBlue = (index - 192) * 4;

		palette[index].peFlags = PC_NOCOLLAPSE;			// set flags
	} // end for index

	// now create the palette object
	if (lpdd->CreatePalette(DDPCAPS_8BIT | DDPCAPS_INITIALIZE | DDPCAPS_ALLOW256,
							 palette, &lpddpal,NULL) != DD_OK)
	   return(0);

	if (lpddsprimary->SetPalette(lpddpal) != DD_OK)		// attach the palette to the primary
	   return(0);

	RECT screen_rect = {0, 0, screen_width, screen_height};	// attach a clipper to the screen
	lpddclipper = DD_Attach_Clipper(lpddsback, 1, &screen_rect);

	if (!Load_Bitmap_File(&bitmap, "ball.bmp"))			// load the 8-bit image
	   return(0);

	// load it's palette into directdraw
	if (FAILED(lpddpal->SetEntries(0, 0, MAX_COLORS_PALETTE, bitmap.palette)))
	   return(0);

	lpddsBall = DDraw_Create_Surface(BLOCK_WIDTH, BLOCK_HEIGHT, 0);		// create surface for ball
	lpddsSpider = DDraw_Create_Surface(BLOCK_WIDTH, BLOCK_HEIGHT, 0);	// create surface for spider
	lpddsHeart = DDraw_Create_Surface(BLOCK_WIDTH, BLOCK_HEIGHT, 0);	// create surface for heart
	lpddsBG = DDraw_Create_Surface(SCREEN_WIDTH, SCREEN_HEIGHT, 0, -1);	// create surface for background
	Scan_Image_Bitmap(&bitmap, lpddsBall,  0, 0);						// scan bitmap into lpddsBall
	Scan_Image_Bitmap(&bitmap, lpddsSpider,  1, 0);						// scan bitmap into lpddsSpider
	Scan_Image_Bitmap(&bitmap, lpddsHeart,  5, 0);						// scan bitmap into lpddsHeart
	for (i = 0; i < NUM_BUG_FRAMES; i++)
	{
		lpddsBug[i] = DDraw_Create_Surface(BLOCK_WIDTH, BLOCK_HEIGHT, 0);	// create surface for bug
		Scan_Image_Bitmap(&bitmap, lpddsBug[i],  i, 1);						// scan bitmap into lpddsBug
	}
	for (i = 0; i < NUM_BONUS_FRAMES; i++)
	{
		lpddsBonus[i] = DDraw_Create_Surface(BLOCK_WIDTH, BLOCK_HEIGHT, 0);	// create surface for bonus
		Scan_Image_Bitmap(&bitmap, lpddsBonus[i],  i, 2);					// scan bitmap into lpddsBonus
	}
	for (i = 0; i < NUM_LIFE_FRAMES; i++)
	{
		lpddsLife[i] = DDraw_Create_Surface(BLOCK_WIDTH, BLOCK_HEIGHT, 0);	// create surface for bonus
		Scan_Image_Bitmap(&bitmap, lpddsLife[i],  i, 3);					// scan bitmap into lpddsLife
	}

	DD_Fill_Surface(lpddsprimary, 0);		// clear out both primary and secondary surfaces
	DD_Fill_Surface(lpddsback, 0);
	DD_Fill_Surface(lpddsBG, 0);

	return(1);		// return success
} // end DD_Init


///////////////////////////////////////////////////////////////

int DD_Shutdown(void)		// this function release all the resources directdraw
{							// allocated, mainly to com objects

	if (lpddclipper)
		lpddclipper->Release();	// release the clipper first

	if (lpddpal)
	   lpddpal->Release();		// release the palette

	if (lpddsback)
		lpddsback->Release();	// release the secondary surface

	if (lpddsprimary)
	   lpddsprimary->Release();	// release the primary surface

	if (lpdd)
		lpdd->Release();		// finally, the main dd object

	Unload_Bitmap_File(&bitmap);

	return(1);					// return success
} // end DD_Shutdown


///////////////////////////////////////////////////////////////   

LPDIRECTDRAWCLIPPER DD_Attach_Clipper(LPDIRECTDRAWSURFACE7 lpdds,
                                      int num_rects,
                                      LPRECT clip_list)
{
	// this function creates a clipper from the sent clip list and attaches
	// it to the sent surface

	int index;                         // looping var
	LPDIRECTDRAWCLIPPER lpddclipper;   // pointer to the newly created dd clipper
	LPRGNDATA region_data;             // pointer to the region data that contains
									   // the header and clip list

	// first create the direct draw clipper
	if ((lpdd->CreateClipper(0,&lpddclipper,NULL))!=DD_OK)
	   return(NULL);

	// now create the clip list from the sent data

	// first allocate memory for region data
	region_data = (LPRGNDATA)malloc(sizeof(RGNDATAHEADER)+num_rects*sizeof(RECT));

	// now copy the rects into region data
	memcpy(region_data->Buffer, clip_list, sizeof(RECT)*num_rects);

	// set up fields of header
	region_data->rdh.dwSize          = sizeof(RGNDATAHEADER);
	region_data->rdh.iType           = RDH_RECTANGLES;
	region_data->rdh.nCount          = num_rects;
	region_data->rdh.nRgnSize        = num_rects*sizeof(RECT);

	region_data->rdh.rcBound.left    =  64000;
	region_data->rdh.rcBound.top     =  64000;
	region_data->rdh.rcBound.right   = -64000;
	region_data->rdh.rcBound.bottom  = -64000;

	for (index = 0; index < num_rects; index++)		// find bounds of all clipping regions
	{
		// test if the next rectangle unioned with the current bound is larger
		if (clip_list[index].left < region_data->rdh.rcBound.left)
		   region_data->rdh.rcBound.left = clip_list[index].left;

		if (clip_list[index].right > region_data->rdh.rcBound.right)
		   region_data->rdh.rcBound.right = clip_list[index].right;

		if (clip_list[index].top < region_data->rdh.rcBound.top)
		   region_data->rdh.rcBound.top = clip_list[index].top;

		if (clip_list[index].bottom > region_data->rdh.rcBound.bottom)
		   region_data->rdh.rcBound.bottom = clip_list[index].bottom;

	} // end for index

	// now we have computed the bounding rectangle region and set up the data
	// now let's set the clipping list

	if ((lpddclipper->SetClipList(region_data, 0))!=DD_OK)
	{
	   free(region_data);		// release memory and return error
	   return(NULL);
	}

	// now attach the clipper to the surface
	if ((lpdds->SetClipper(lpddclipper))!=DD_OK)
	{
	   free(region_data);		// release memory and return error
	   return(NULL);
	}

	// all is well, so release memory and send back the pointer to the new clipper
	free(region_data);

	return(lpddclipper);
} // end DD_Attach_Clipper


///////////////////////////////////////////////////////////////
   
int DD_Flip(void)		// this function flip the primary surface with the secondary surface
{
	while(lpddsprimary->Flip(NULL, DDFLIP_WAIT)!=DD_OK);	// flip the surface
	return(1);		// return success

} // end DD_Flip


///////////////////////////////////////////////////////////////

DWORD Start_Clock(void)		// this function starts the clock, that is, saves the current
{							// count, use in conjunction with Wait_Clock()

	return(start_clock_count = Get_Clock());

} // end Start_Clock


///////////////////////////////////////////////////////////////

DWORD Get_Clock(void)		// this function returns the current tick count
{

	return(GetTickCount());		// return time

} // end Get_Clock


///////////////////////////////////////////////////////////////

DWORD Wait_Clock(DWORD count)	// this function is used to wait for a specific number of clicks
{								// since the call to Start_Clock
	
	while((Get_Clock() - start_clock_count) < count);
		return(Get_Clock());

} // end Wait_Clock


///////////////////////////////////////////////////////////////

int DD_Fill_Surface(LPDIRECTDRAWSURFACE7 lpdds,int color)
{
	DDBLTFX ddbltfx; // this contains the DDBLTFX structure

	DD_INIT_STRUCT(ddbltfx);		// clear out the structure and set the size field
	ddbltfx.dwFillColor = color;	// set the dwfillcolor field to the desired color

	// ready to blt to surface
	lpdds->Blt(NULL,       // ptr to dest rectangle
			   NULL,       // ptr to source surface, NA            
			   NULL,       // ptr to source rectangle, NA
			   DDBLT_COLORFILL | DDBLT_WAIT | DDBLT_ASYNC,   // fill and wait                   
			   &ddbltfx);  // ptr to DDBLTFX structure

	return(1);		// return success
} // end DD_Fill_Surface


///////////////////////////////////////////////////////////////   

int Draw_Rectangle(int x1, int y1, int x2, int y2, int color,
                   LPDIRECTDRAWSURFACE7 lpdds)
{			// this function uses directdraw to draw a filled rectangle

	DDBLTFX ddbltfx; // this contains the DDBLTFX structure
	RECT fill_area;  // this contains the destination rectangle

	DD_INIT_STRUCT(ddbltfx);		// clear out the structure and set the size field 
	ddbltfx.dwFillColor = color;	// set the dwfillcolor field to the desired color

	fill_area.top    = y1;			// fill in the destination rectangle data (your data)
	fill_area.left   = x1;
	fill_area.bottom = y2;
	fill_area.right  = x2;

	// ready to blt to surface, in this case blt to primary
	lpdds->Blt(&fill_area, // ptr to dest rectangle
			   NULL,       // ptr to source surface, NA            
			   NULL,       // ptr to source rectangle, NA
			   DDBLT_COLORFILL | DDBLT_WAIT | DDBLT_ASYNC,   // fill and wait                   
			   &ddbltfx);  // ptr to DDBLTFX structure
	
	return(1);		// return success

} // end Draw_Rectangle


///////////////////////////////////////////////////////////////

int Draw_Text_GDI(char *text, int x,int y,int color, LPDIRECTDRAWSURFACE7 lpdds)
{
	// this function draws the sent text on the sent surface 
	// using color index as the color in the palette

	HDC xdc; // the working dc

	if (lpdds->GetDC(&xdc)!=DD_OK)	// get the dc from surface
	   return(0);

	// set the colors for the text up
	SetTextColor(xdc,RGB(palette[color].peRed,palette[color].peGreen,palette[color].peBlue) );

	SetBkMode(xdc, TRANSPARENT);	// set background mode to transparent so black isn't copied
	TextOut(xdc,x,y,text,strlen(text));		// draw the text a
	lpdds->ReleaseDC(xdc);			// release the dc

	return(1);						// return success

} // end Draw_Text_GDI


///////////////////////////////////////////////////////////////

int Load_Bitmap_File(BITMAP_FILE_PTR bitmap, char *filename)
{
	// this function opens a bitmap file and loads the data into bitmap

	int file_handle,				// the file handle
		index;						// looping index

	UCHAR   *temp_buffer = NULL;	// used to convert 24 bit images to 16 bit
	OFSTRUCT file_data;				// the file data information

	// open the file if it exists
	if ((file_handle = OpenFile(filename, &file_data, OF_READ))==-1)
	   return(0);

	// now load the bitmap file header
	_lread(file_handle, &bitmap->bitmapfileheader, sizeof(BITMAPFILEHEADER));

	// test if this is a bitmap file
	if (bitmap->bitmapfileheader.bfType != BITMAP_ID)
	{
	   _lclose(file_handle);	// close the file
	   return(0);				// return error
	}

	// now we know this is a bitmap, so read in all the sections
	// first the bitmap infoheader
	_lread(file_handle, &bitmap->bitmapinfoheader, sizeof(BITMAPINFOHEADER));

	// now load the color palette if there is one
	if (bitmap->bitmapinfoheader.biBitCount == 8)
	{
		_lread(file_handle, &bitmap->palette, MAX_COLORS_PALETTE * sizeof(PALETTEENTRY));

		// now set all the flags in the palette correctly and fix the reversed BGR RGBQUAD data format
		for (index = 0; index < MAX_COLORS_PALETTE; index++)
		{
		   // reverse the red and green fields
		   int temp_color                = bitmap->palette[index].peRed;
		   bitmap->palette[index].peRed  = bitmap->palette[index].peBlue;
		   bitmap->palette[index].peBlue = temp_color;

		   bitmap->palette[index].peFlags = PC_NOCOLLAPSE;	// always set the flags word to this
		}
	}

	// finally the image data itself
	_lseek(file_handle, -(int)(bitmap->bitmapinfoheader.biSizeImage), SEEK_END);

	// now read in the image, if the image is 8 or 16 bit then simply read it
	// but if its 24 bit then read it into a temporary area and then convert
	// it to a 16 bit image

	if (bitmap->bitmapinfoheader.biBitCount == 8 || bitmap->bitmapinfoheader.biBitCount == 16 || 
		bitmap->bitmapinfoheader.biBitCount == 24)
	{
		if (bitmap->buffer)			// delete the last image if there was one
		   free(bitmap->buffer);

		// allocate the memory for the image
		if (!(bitmap->buffer = (UCHAR *)malloc(bitmap->bitmapinfoheader.biSizeImage)))
		{
			_lclose(file_handle);	// close the file
			return(0);				// return error
		} // end if

	   // now read it in
	   _lread(file_handle, bitmap->buffer,bitmap->bitmapinfoheader.biSizeImage);
	}
	else
	{
	   return(0);					// serious problem
	}

	_lclose(file_handle);			// close the file

	// flip the bitmap
	Flip_Bitmap(bitmap->buffer, 
				bitmap->bitmapinfoheader.biWidth * (bitmap->bitmapinfoheader.biBitCount / 8), 
				bitmap->bitmapinfoheader.biHeight);

	return(1);					// return success

} // end Load_Bitmap_File


///////////////////////////////////////////////////////////

int Unload_Bitmap_File(BITMAP_FILE_PTR bitmap)
{
	// this function releases all memory associated with "bitmap"
	if (bitmap->buffer)
	{
	   free(bitmap->buffer);	// release memory
	   bitmap->buffer = NULL;	// reset pointer

	} // end if

	return(1);					// return success

} // end Unload_Bitmap_File


///////////////////////////////////////////////////////////

int Flip_Bitmap(UCHAR *image, int bytes_per_line, int height)
{
	// this function is used to flip bottom-up .BMP images

	UCHAR *buffer; // used to perform the image processing
	int index;     // looping index

	if (!(buffer = (UCHAR *)malloc(bytes_per_line * height)))	// allocate the temporary buffer
	   return(0);

	memcpy(buffer, image, bytes_per_line * height);				// copy image to work area

	for (index = 0; index < height; index++)					// flip vertically
		memcpy(&image[((height - 1) - index) * bytes_per_line],
			   &buffer[index * bytes_per_line], bytes_per_line);

	free(buffer);		// release the memory

	return(1);			// return success

} // end Flip_Bitmap


///////////////////////////////////////////////////////////////

LPDIRECTDRAWSURFACE7 DDraw_Create_Surface(int width, int height, int mem_flags, int color_key)
{
	// this function creates an offscreen plain surface

	DDSURFACEDESC2 ddsd;			// working description
	LPDIRECTDRAWSURFACE7 lpdds;		// temporary surface
    
	// set to access caps, width, and height
	memset(&ddsd, 0, sizeof(ddsd));
	ddsd.dwSize  = sizeof(ddsd);
	ddsd.dwFlags = DDSD_CAPS | DDSD_WIDTH | DDSD_HEIGHT;

	// set dimensions of the new bitmap surface
	ddsd.dwWidth  =  width;
	ddsd.dwHeight =  height;

	// set surface to offscreen plain
	ddsd.ddsCaps.dwCaps = DDSCAPS_OFFSCREENPLAIN | mem_flags;

	if (FAILED(lpdd->CreateSurface(&ddsd, &lpdds, NULL)))		// create the surface
	   return(NULL);

	if (color_key >= 0)							// test if user wants a color key
	{
	   // set color key to color 0
	   DDCOLORKEY color_key;					// used to set color key
	   color_key.dwColorSpaceLowValue  = 0;
	   color_key.dwColorSpaceHighValue = 0;

	   // now set the color key for source blitting
	   lpdds->SetColorKey(DDCKEY_SRCBLT, &color_key);
	}

	return(lpdds);		// return surface

} // end DDraw_Create_Surface


///////////////////////////////////////////////////////////////

int DDraw_Draw_Surface(LPDIRECTDRAWSURFACE7 source, // source surface to draw
                      int x, int y,                 // position to draw at
                      int width, int height,        // size of source surface
                      LPDIRECTDRAWSURFACE7 dest,    // surface to draw the surface on
                      int transparent)				// transparency flag
{
	// draw a bob at the x,y defined in the BOB
	// on the destination surface defined in dest

	RECT dest_rect,   // the destination rectangle
		 source_rect; // the source rectangle                             

	// fill in the destination rect
	dest_rect.left   = x;
	dest_rect.top    = y;
	dest_rect.right  = x + width;
	dest_rect.bottom = y + height;

	// fill in the source rect
	source_rect.left    = 0;
	source_rect.top     = 0;
	source_rect.right   = width;
	source_rect.bottom  = height;

	// test transparency flag
	if (transparent)						// enable color key blit
	{
	   if (FAILED(dest->Blt(&dest_rect, source,
						 &source_rect, (DDBLT_WAIT | DDBLT_KEYSRC),
						 NULL)))			// blt to destination surface
			return(0);
	}
	else									// perform blit without color key
	{
	   if (FAILED(dest->Blt(&dest_rect, source,
						 &source_rect, (DDBLT_WAIT),
						 NULL)))
			return(0);

	}

	return(1);		// return success

} // end DDraw_Draw_Surface


///////////////////////////////////////////////////////////////

int Scan_Image_Bitmap(BITMAP_FILE_PTR bitmap,     // bitmap file to scan image data from
                      LPDIRECTDRAWSURFACE7 lpdds, // surface to hold data
                      int cx, int cy)             // cell to scan image from
{
	// this function extracts a bitmap out of a bitmap file

	UCHAR *source_ptr,   // working pointers
		  *dest_ptr;

	DDSURFACEDESC2 ddsd;  //  direct draw surface description 

	// get the addr to destination surface memory
	ddsd.dwSize = sizeof(ddsd);			// set size of the structure
	lpdds->Lock(NULL,					// lock the display surface
				&ddsd,
				DDLOCK_WAIT | DDLOCK_SURFACEMEMORYPTR,
				NULL);

	cx = cx * (ddsd.dwWidth + 1) + 1;	// compute position to start scanning bits from
	cy = cy * (ddsd.dwHeight + 1) + 1;

	// extract bitmap data
	source_ptr = bitmap->buffer + cy * bitmap->bitmapinfoheader.biWidth + cx;

	dest_ptr = (UCHAR *)ddsd.lpSurface;	// assign a pointer to the memory surface for manipulation

	// iterate thru each scanline and copy bitmap
	for (WORD index_y = 0; index_y < ddsd.dwHeight; index_y++)
	{
		memcpy(dest_ptr, source_ptr, ddsd.dwWidth);	// copy next line of data to destination

		dest_ptr   += (ddsd.lPitch);				// advance pointers
		source_ptr += bitmap->bitmapinfoheader.biWidth;
	}

	lpdds->Unlock(NULL);			// unlock the surface 

	return(1);						// return success

} // end Scan_Image_Bitmap

