// palmanip.cpp - palette manipulation functions from LaMothe's game library

#include <ddraw.h>								// directX includes
#include <stdio.h>
#include "palmanip.h"
#include "global.h"


extern PALETTEENTRY		palette[256];			// color palette
extern LPDIRECTDRAWPALETTE   lpddpal;			// a pointer to the created dd palette


int Save_Palette(LPPALETTEENTRY sav_palette)	// this function saves the current palette 
{

	memcpy(sav_palette, palette, MAX_COLORS_PALETTE * sizeof(PALETTEENTRY));
	return(1);	// return success

} // end Save_Palette


///////////////////////////////////////////////////////////////

int Set_Palette(LPPALETTEENTRY set_palette)		// this function writes the sent palette
{
	// first save the new palette in shadow
	memcpy(palette, set_palette, MAX_COLORS_PALETTE * sizeof(PALETTEENTRY));
	lpddpal->SetEntries(0, 0, MAX_COLORS_PALETTE, palette);		// now set the new palette

	return(1);		// return success
} // end Set_Palette


///////////////////////////////////////////////////////////////

void DarkenScreen()		// this function gradually darkens screen
{
	int pal_reg;				// used as loop counter
	PALETTEENTRY color;         // temporary color
	PALETTEENTRY work_palette[MAX_COLORS_PALETTE];  // used as a working palette

	for (int index = 0; index < 60; index++)
	{
		Save_Palette(work_palette);// get the palette 

		for (pal_reg=1; pal_reg<MAX_COLORS_PALETTE; pal_reg++)		// process each color
		{
			color = work_palette[pal_reg];							// get the entry data

			// test if this color register is already black
			if (color.peRed > 4)
				color.peRed -= 3;
			else
				color.peRed = 0;

			if (color.peGreen > 4)
				color.peGreen -= 3;
			else
				color.peGreen = 0;

			if (color.peBlue > 4)
				color.peBlue -= 3;
			else
				color.peBlue = 0;

			work_palette[pal_reg] = color;	// set the color to a diminished intensity
		} // end for pal_reg
		
		Set_Palette(work_palette);			// write the palette back out

		Sleep(15);		// wait a bit

	} // end for index
}


///////////////////////////////////////////////////////////////

int Save_Palette_To_File(char *filename, LPPALETTEENTRY palette)
{
	// this function saves a palette to disk

	FILE *fp_file;	// working file
	
	if ((fp_file = fopen(filename, "w"))==NULL)					// try and open file
	   return(0);

	for (int index = 0; index < MAX_COLORS_PALETTE; index++)	// write in all 256 colors RGBF
    {
		fprintf(fp_file, "\n%d   %d %d %d %d",
									index,
									palette[index].peRed,
                                    palette[index].peGreen,
                                    palette[index].peBlue,
                                    palette[index].peFlags);
    }

	fclose(fp_file);	// close the file

	return(1);			// return success

} // end Save_Palette_To_Disk


