// xonix.h - header file for the main game file xonix.cpp

#ifndef XONIX_H
#define XONIX_H


// INCLUDES ///////////////////////////////////////////////////

#define WIN32_LEAN_AND_MEAN // include all macros
#define INITGUID            // include all GUIDs 

#include <windows.h>        // include important windows stuff
#include <windowsx.h> 
#include <mmsystem.h>

#include <iostream.h>       // include important C/C++ stuff
#include <conio.h>
#include <stdlib.h>
#include <malloc.h>
#include <memory.h>
#include <string.h>
#include <stdarg.h>
#include <stdio.h>
#include <math.h>
#include <io.h>
#include <fcntl.h>
#include <fstream.h>

#include <ddraw.h>  // directX includes
#include <dsound.h>
#include <dmksctrl.h>
#include <dmusici.h>
#include <dmusicc.h>
#include <dmusicf.h>

#include "ddstuff.h"       // game library includes
#include "filling.h"
#include "global.h"
#include "palmanip.h"
#include "soundlib.h"

// DEFINES ////////////////////////////////////////////////////

// defines for windows 
#define WINDOW_CLASS_NAME "WIN3DCLASS"  // class name

// states for game loop
#define GAME_STATE_INIT         0
#define GAME_STATE_START_LEVEL  1
#define GAME_STATE_RUN          2
#define GAME_STATE_SHUTDOWN     3
#define GAME_STATE_EXIT         4
#define GAME_STATE_WIN			5
#define GAME_STATE_LOSE			6
#define GAME_STATE_PAUSE		7

// enemy defines
#define BUG_START_X			(NUM_BLOCK_COLUMNS / 2)
#define BUG_START_Y			(NUM_BLOCK_ROWS / 2)

// sound effects defines
#define DEATH_SOUND				0
#define WIN_SOUND				1
#define BONUS_SOUND				2
#define NUM_SOUNDS				3


// PROTOTYPES /////////////////////////////////////////////////

// game console
int Game_Init(void *parms = NULL);
int Game_Shutdown(void *parms = NULL);
int Game_Main(void *parms = NULL);

// GLOBALS ////////////////////////////////////////////////////

HWND main_window_handle  = NULL; // save the window handle
HINSTANCE main_instance  = NULL; // save the instance
int game_state           = GAME_STATE_INIT; // starting state

int ball_x = 0, ball_y = 0;			// tracks position of ball
int ballXOrig = 0, ballYOrig = 0;	// store the original position of ball before drawing trail
int x, y;							// current screen coordinates of ball
int score = 0;						// the score
int level = 1;						// the current level
int numOccupied = 0;				// tracks number of blocks occupied
int percentOccupied = 0;			// the percentage of occupied blocks
int frames = 0;						// counts the # of frames within 1 second (fps)
int fps = 0;						// holds the fps
DWORD startTime = 0;				// time when counting fps begins
bool trail = false;					// indicates if the ball is leaving trail
bool flipped = false;				// indicates if the serfaces were flipped
int  idleTime = 35;					// delay between drawing frames in milliseconds
int  numBugs = 1, numSpiders = 1;	// number of bugs and spiders on the current level
int  moveCounter;					// variable to keep tack of the number of moves
bool bonusOn = false;				// indicates if bonus has been put on the screen
bool lifeOn = false;				// indicates if life bonus has been put on the screen
int  midiID[NUM_LEVELS + 2];		// storage for song IDs
int  sound[NUM_SOUNDS];				// storage for digital sounds
char buffer[80];					// used to print text
int  livesLeft = 5;					// five lives are given to the player
int  initLivesLeft = livesLeft;		// number of lives at the beginning of the level
int  initScore = score;				// the score at the beginning of the level

struct anim_struct bug[NUM_LEVELS];
struct anim_struct spider[NUM_LEVELS];
struct anim_struct bonus;
struct anim_struct life;
struct anim_struct heart;


BITMAP_FILE				bitmap;		// holds the bitmap
LPDIRECTDRAWSURFACE7	lpddsBall;
LPDIRECTDRAWSURFACE7	lpddsSpider;
LPDIRECTDRAWSURFACE7	lpddsBG;
LPDIRECTDRAWSURFACE7	lpddsHeart;
LPDIRECTDRAWSURFACE7	lpddsBug[NUM_BUG_FRAMES];
LPDIRECTDRAWSURFACE7	lpddsBonus[NUM_BONUS_FRAMES];
LPDIRECTDRAWSURFACE7	lpddsLife[NUM_LIFE_FRAMES];

UCHAR blocks[NUM_BLOCK_COLUMNS][NUM_BLOCK_ROWS]; // this contains the game grid data  

ofstream outFile; // debug info output file

#endif